import os
import uvicorn
import importlib.resources

from .function import Function, _get_api_key, _get_base_url
from .web import web, Config

CYCLS_PATH = importlib.resources.files('cycls')

THEMES = ["default", "dev"]


class App(Function):
    """App extends Function with web UI serving capabilities."""

    def __init__(self, func, name, theme="default", pip=None, apt=None, run_commands=None, copy=None, copy_public=None,
                 auth=False, header=None, intro=None, title=None, plan="free", analytics=False,
                 memory="1Gi", force_rebuild=False):
        if theme not in THEMES:
            raise ValueError(f"Unknown theme: {theme}. Available: {THEMES}")
        self.user_func = func
        self.theme = theme
        self.copy_public = copy_public or []
        self.memory = memory

        self.config = Config(
            header=header,
            intro=intro,
            title=title,
            auth=auth,
            plan=plan,
            analytics=analytics,

        )

        # Build files dict for Function (theme is inside cycls/)
        files = {str(CYCLS_PATH): "cycls"}
        files.update({f: f for f in copy or []})
        files.update({f: f"public/{f}" for f in self.copy_public})

        # Fetch theme from GitHub releases during container build
        theme_cmd = "mkdir -p /app/cycls/themes/default && cd /app/cycls/themes/default && curl -fsSLO https://github.com/Cycls/agentUI/releases/download/latest/agentUI.zip && unzip -o agentUI.zip && rm agentUI.zip"
        all_run_commands = [theme_cmd, *(run_commands or [])]

        super().__init__(
            func=func,
            name=name,
            pip=["fastapi[standard]", "pyjwt", "cryptography", "uvicorn", "python-dotenv", "docker", *(pip or [])],
            apt=["curl", "unzip", *(apt or [])],
            run_commands=all_run_commands,
            copy=files,
            base_url=_get_base_url(),
            api_key=_get_api_key(),
            force_rebuild=force_rebuild
        )

    def __call__(self, *args, **kwargs):
        return self.user_func(*args, **kwargs)

    def _prepare_func(self, prod):
        self.config.set_prod(prod)
        self.config.public_path = f"cycls/themes/{self.theme}"
        user_func, config, name = self.user_func, self.config, self.name
        self.func = lambda port: __import__("cycls").web.serve(user_func, config, name, port)

    def _local(self, port=8080):
        """Run directly with uvicorn (no Docker)."""
        print(f"Starting local server at localhost:{port}")
        self.config.public_path = str(CYCLS_PATH.joinpath(f"themes/{self.theme}"))
        self.config.set_prod(False)
        uvicorn.run(web(self.user_func, self.config), host="0.0.0.0", port=port)

    def local(self, port=8080, watch=True):
        """Run locally in Docker with file watching by default."""
        if os.environ.get('_CYCLS_WATCH'):
            watch = False
        self._prepare_func(prod=False)
        self.watch(port=port) if watch else self.run(port=port)

    def deploy(self, port=8080, memory=None):
        """Deploy to production."""
        if self.api_key is None:
            raise RuntimeError("Missing API key. Set cycls.api_key or CYCLS_API_KEY environment variable.")
        self._prepare_func(prod=True)
        return super().deploy(port=port, memory=memory or self.memory)


def app(name=None, **kwargs):
    """Decorator that transforms a function into a deployable App."""
    if kwargs.get("plan") == "cycls_pass":
        kwargs["auth"] = True
        kwargs["analytics"] = True

    def decorator(func):
        return App(func=func, name=name or func.__name__, **kwargs)
    return decorator



#!/usr/bin/env python3
"""cycls chat - Claude Code style CLI for cycls agents"""

import json, os, re, sys
import httpx

RESET, BOLD, DIM = "\033[0m", "\033[1m", "\033[2m"
BLUE, GREEN, YELLOW, RED = "\033[34m", "\033[32m", "\033[33m", "\033[31m"
CALLOUTS = {"success": ("✓", GREEN), "warning": ("⚠", YELLOW), "info": ("ℹ", BLUE), "error": ("✗", RED)}

separator = lambda: f"{DIM}{'─' * min(os.get_terminal_size().columns if sys.stdout.isatty() else 80, 80)}{RESET}"
markdown = lambda text: re.sub(r"\*\*(.+?)\*\*", f"{BOLD}\\1{RESET}", text)
header = lambda title, meta, color=GREEN, dim=False: print(f"{color}●{RESET} {BOLD}{title}{RESET}\n  ⎿  {meta}{DIM if dim else ''}", flush=True)


def table(headers, rows):
    if not headers: return
    widths = [max(len(str(h)), *(len(str(r[i])) for r in rows if i < len(r))) for i, h in enumerate(headers)]
    line = lambda left, mid, right: left + mid.join("─" * (w + 2) for w in widths) + right
    row = lambda cells, bold=False: "│" + "│".join(f" {BOLD if bold else ''}{str(cells[i] if i < len(cells) else '').ljust(widths[i])}{RESET if bold else ''} " for i in range(len(widths))) + "│"
    print(f"{line('┌', '┬', '┐')}\n{row(headers, True)}\n{line('├', '┼', '┤')}")
    for r in rows: print(row(r))
    print(line("└", "┴", "┘"))


def chat(url):
    messages, endpoint = [], f"{url.rstrip('/')}/chat/cycls"
    print(f"\n{BOLD}cycls{RESET} {DIM}|{RESET} {url}\n")

    while True:
        try:
            print(separator())
            user_input = input(f"{BOLD}{BLUE}❯{RESET} ").strip()
            print(separator())

            if not user_input: continue
            if user_input in ("/q", "exit", "quit"): break
            if user_input == "/c": messages, _ = [], print(f"{GREEN}⏺ Cleared{RESET}"); continue

            messages.append({"role": "user", "content": user_input})
            block, tbl = None, ([], [])

            def close():
                nonlocal block, tbl
                if block == "thinking": print(RESET)
                if block == "text": print()
                if block == "table" and tbl[0]: table(*tbl); tbl = ([], [])
                if block: print()
                block = None

            with httpx.stream("POST", endpoint, json={"messages": messages}, timeout=None) as response:
                for line in response.iter_lines():
                    if not line.startswith("data: ") or line == "data: [DONE]": continue
                    data = json.loads(line[6:])
                    type = data.get("type")

                    if type is None:
                        print(markdown(data if isinstance(data, str) else data.get("text", "")), end="", flush=True); continue

                    if type != block: close()

                    if type in ("thinking", "text"):
                        if block != type: header(type.capitalize(), "Live", dim=(type == "thinking")); block = type
                        print((markdown if type == "text" else str)(data.get(type, "")), end="", flush=True)
                    elif type == "code":
                        code = data.get("code", ""); header(f"Code({data.get('language', '')})", f"{code.count(chr(10))+1} lines"); print(code, flush=True); block = type
                    elif type == "status":
                        print(f"{DIM}[{data.get('status', '')}]{RESET} ", end="", flush=True)
                    elif type == "table":
                        if "headers" in data:
                            if tbl[0]: table(*tbl)
                            header("Table", f"{len(data['headers'])} cols"); tbl, block = (data["headers"], []), type
                        elif "row" in data: tbl[1].append(data["row"])
                    elif type == "callout":
                        style = data.get("style", "info"); icon, color = CALLOUTS.get(style, ("•", RESET))
                        header(style.capitalize(), f"{icon} {data.get('callout', '')}", color=color); block = type
                    elif type == "image":
                        header("Image", data.get("src", "")); block = type

            close()

        except KeyboardInterrupt: print()
        except EOFError: break
        except (httpx.ReadError, httpx.ConnectError) as e: print(f"{RED}⏺ Connection error: {e}{RESET}"); messages and messages.pop()


def main():
    if len(sys.argv) < 3:
        print("Usage: cycls chat <url|port>")
        sys.exit(1)
    arg = sys.argv[2]
    if arg.isdigit():
        port = int(arg)
        if not (1 <= port <= 65535):
            print(f"Error: Invalid port {port}. Must be between 1 and 65535.")
            sys.exit(1)
        url = f"http://localhost:{port}"
    else:
        url = arg
    chat(url)


if __name__ == "__main__":
    main()



import json, inspect, secrets, os
from datetime import datetime, timezone
from urllib.parse import quote
from pathlib import Path
from pydantic import BaseModel
from typing import Optional, Any
from .auth import PK_LIVE, PK_TEST, JWKS_PROD, JWKS_TEST

class Config(BaseModel):
    public_path: str = "theme"
    header: Optional[str] = None
    intro: Optional[str] = None
    title: Optional[str] = None
    prod: bool = False
    auth: bool = False
    plan: str = "free"
    analytics: bool = False
    pk: Optional[str] = None
    jwks: Optional[str] = None

    def set_prod(self, prod: bool):
        self.prod = prod
        self.pk = PK_LIVE if prod else PK_TEST
        self.jwks = JWKS_PROD if prod else JWKS_TEST

async def openai_encoder(stream):
    if inspect.isasyncgen(stream):
        async for msg in stream:
            if msg: yield f"data: {json.dumps({'choices': [{'delta': {'content': msg}}]})}\n\n"
    else:
        for msg in stream:
            if msg: yield f"data: {json.dumps({'choices': [{'delta': {'content': msg}}]})}\n\n"
    yield "data: [DONE]\n\n"

def sse(item):
    if not item: return None
    if not isinstance(item, dict): item = {"type": "text", "text": item}
    return f"data: {json.dumps(item)}\n\n"

async def encoder(stream):
    if inspect.isasyncgen(stream):
        async for item in stream:
            if msg := sse(item): yield msg
    else:
        for item in stream:
            if msg := sse(item): yield msg
    yield "data: [DONE]\n\n"

class Messages(list):
    """A list that provides text-only messages by default, with .raw for full data."""
    def __init__(self, raw_messages):
        self._raw = raw_messages
        text_messages = []
        for m in raw_messages:
            text_content = "".join(
                p.get("text", "") for p in m.get("parts", []) if p.get("type") == "text"
            )
            text_messages.append({
                "role": m.get("role"),
                "content": m.get("content") or text_content
            })
        super().__init__(text_messages)

    @property
    def raw(self):
        return self._raw

def web(func, config):
    from fastapi import FastAPI, Request, HTTPException, status, Depends, UploadFile, File
    from fastapi.responses import StreamingResponse, FileResponse
    from fastapi.security import HTTPBearer, HTTPAuthorizationCredentials
    import jwt
    from jwt import PyJWKClient
    from pydantic import EmailStr
    from typing import List, Optional, Any
    from fastapi.staticfiles import StaticFiles

    if isinstance(config, dict):
        config = Config(**config)

    jwks = PyJWKClient(config.jwks)

    class User(BaseModel):
        id: str
        name: Optional[str] = None
        email: EmailStr
        org_id: Optional[str] = None
        org_name: Optional[str] = None
        org_slug: Optional[str] = None
        plan_name: Optional[str] = None
        plan_id: Optional[str] = None
        plan_slug: Optional[str] = None

    class Context(BaseModel):
        messages: Any
        user: Optional[User] = None

        model_config = {"arbitrary_types_allowed": True}

        @property
        def last_message(self) -> str:
            if self.messages:
                return self.messages[-1].get("content", "")
            return ""

    app = FastAPI()
    bearer_scheme = HTTPBearer()

    def validate(bearer: HTTPAuthorizationCredentials = Depends(bearer_scheme)):
        try:
            key = jwks.get_signing_key_from_jwt(bearer.credentials)
            decoded = jwt.decode(bearer.credentials, key.key, algorithms=["RS256"], leeway=10)
            return {"type": "user",
                    "user": {"id": decoded.get("id"), "name": decoded.get("name"), "email": decoded.get("email"),
                             "org_id": decoded.get("org_id"), "org_name": decoded.get("org_name"), "org_slug": decoded.get("org_slug")}}
        except jwt.ExpiredSignatureError:
            raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail="Token expired", headers={"WWW-Authenticate": "Bearer"})
        except jwt.InvalidTokenError as e:
            raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail=f"Invalid token: {e}", headers={"WWW-Authenticate": "Bearer"})
        except Exception as e:
            raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail=f"Auth error: {e}", headers={"WWW-Authenticate": "Bearer"})

    auth = Depends(validate) if config.auth else None

    @app.post("/")
    @app.post("/chat/cycls")
    @app.post("/chat/completions")
    async def back(request: Request, jwt: Optional[dict] = auth):
        data = await request.json()
        messages = data.get("messages")
        user_data = jwt.get("user") if jwt else None
        if user_data and data.get("plan"):
            user_data.update(data["plan"])
        user = User(**user_data) if user_data else None

        context = Context(messages=Messages(messages), user=user)
        stream = await func(context) if inspect.iscoroutinefunction(func) else func(context)

        if request.url.path == "/chat/completions":
            stream = openai_encoder(stream)
        elif request.url.path == "/chat/cycls":
            stream = encoder(stream)
        return StreamingResponse(stream, media_type="text/event-stream")

    @app.get("/config")
    async def get_config():
        return config

    @app.post("/attachments")
    async def upload_attachment(request: Request, file: UploadFile = File(...), jwt: Optional[dict] = auth):
        token = secrets.token_urlsafe(32)
        token_dir = Path(f"/workspace/attachments/{token}")
        token_dir.mkdir(parents=True, exist_ok=True)

        filename = Path(file.filename).name
        file_path = (token_dir / filename).resolve()
        if not file_path.is_relative_to(token_dir.resolve()):
            raise HTTPException(status_code=400, detail="Invalid filename")
        with open(file_path, "wb") as f:
            f.write(await file.read())

        return {"url": f"/attachments/{token}/{quote(filename)}"}

    @app.get("/attachments/{token}/{filename}")
    async def get_attachment(token: str, filename: str):
        token_dir = Path(f"/workspace/attachments/{token}")
        file_path = (token_dir / Path(filename).name).resolve()
        if not file_path.is_relative_to(token_dir.resolve()):
            raise HTTPException(status_code=404, detail="File not found")

        if not file_path.exists():
            raise HTTPException(status_code=404, detail="File not found")

        return FileResponse(file_path)

    # ---- Helper ----

    def user_root(jwt):
        user = jwt["user"]
        return Path(f"/workspace/{user['org_id']}") if user.get("org_id") else Path(f"/workspace/{user['id']}")

    # ---- Sessions API ----

    @app.get("/sessions")
    async def list_sessions(jwt: dict = auth):
        sessions_dir = user_root(jwt) / ".sessions" / jwt["user"]["id"]
        if not sessions_dir.is_dir():
            return []
        items = []
        for f in sessions_dir.iterdir():
            if f.suffix != ".json":
                continue
            try:
                data = json.loads(f.read_text())
                items.append({"id": data.get("id", f.stem), "title": data.get("title", ""), "updatedAt": data.get("updatedAt", "")})
            except (json.JSONDecodeError, OSError):
                continue
        items.sort(key=lambda s: s.get("updatedAt", ""), reverse=True)
        return items

    @app.get("/sessions/{session_id}")
    async def get_session(session_id: str, jwt: dict = auth):
        session_file = user_root(jwt) / ".sessions" / jwt["user"]["id"] / f"{session_id}.json"
        if not session_file.is_file():
            raise HTTPException(status_code=404, detail="Session not found")
        return json.loads(session_file.read_text())

    @app.put("/sessions/{session_id}")
    async def put_session(session_id: str, request: Request, jwt: dict = auth):
        sessions_dir = user_root(jwt) / ".sessions" / jwt["user"]["id"]
        sessions_dir.mkdir(parents=True, exist_ok=True)
        data = await request.json()
        data["id"] = session_id
        data["updatedAt"] = datetime.now(timezone.utc).isoformat()
        if "createdAt" not in data:
            data["createdAt"] = data["updatedAt"]
        session_file = sessions_dir / f"{session_id}.json"
        session_file.write_text(json.dumps(data))
        return data

    @app.delete("/sessions/{session_id}")
    async def delete_session(session_id: str, jwt: dict = auth):
        session_file = user_root(jwt) / ".sessions" / jwt["user"]["id"] / f"{session_id}.json"
        if not session_file.is_file():
            raise HTTPException(status_code=404, detail="Session not found")
        session_file.unlink()
        return {"ok": True}

    # ---- File API ----

    def safe_path(jwt, rel: str) -> Path:
        root = user_root(jwt)
        resolved = (root / rel).resolve()
        if not resolved.is_relative_to(root.resolve()):
            raise HTTPException(status_code=403, detail="Path traversal denied")
        return resolved

    @app.get("/files")
    async def list_files(request: Request, jwt: dict = auth):
        subpath = request.query_params.get("path", "")
        target = safe_path(jwt, subpath)
        if not target.is_dir():
            raise HTTPException(status_code=404, detail="Directory not found")
        items = []
        for entry in os.scandir(target):
            if entry.name.startswith("."):
                continue
            stat = entry.stat()
            items.append({
                "name": entry.name,
                "type": "directory" if entry.is_dir() else "file",
                "size": stat.st_size,
                "modified": datetime.fromtimestamp(stat.st_mtime, tz=timezone.utc).isoformat(),
            })
        items.sort(key=lambda f: f["name"])
        return items

    @app.get("/files/{path:path}")
    async def get_file(path: str, request: Request, jwt: dict = auth):
        file_path = safe_path(jwt, path)
        if not file_path.is_file():
            raise HTTPException(status_code=404, detail="File not found")
        if request.query_params.get("download") is not None:
            return FileResponse(file_path, filename=file_path.name)
        return FileResponse(file_path)

    @app.put("/files/{path:path}")
    async def put_file(path: str, request: Request, file: UploadFile = File(...), jwt: dict = auth):
        file_path = safe_path(jwt, path)
        file_path.parent.mkdir(parents=True, exist_ok=True)
        file_path.write_bytes(await file.read())
        return {"ok": True}

    @app.patch("/files/{path:path}")
    async def rename(path: str, request: Request, jwt: dict = auth):
        src = safe_path(jwt, path)
        if not src.exists():
            raise HTTPException(status_code=404, detail="Not found")
        data = await request.json()
        dest = safe_path(jwt, data["to"])
        dest.parent.mkdir(parents=True, exist_ok=True)
        src.rename(dest)
        return {"ok": True}

    @app.post("/files/{path:path}")
    async def mkdir(path: str, jwt: dict = auth):
        dir_path = safe_path(jwt, path)
        dir_path.mkdir(parents=True, exist_ok=True)
        return {"ok": True}

    @app.delete("/files/{path:path}")
    async def delete_path(path: str, jwt: dict = auth):
        import shutil
        target = safe_path(jwt, path)
        if not target.exists():
            raise HTTPException(status_code=404, detail="Not found")
        if target.is_dir():
            shutil.rmtree(target)
        else:
            target.unlink()
        return {"ok": True}

    # ---- Static mounts (must be last) ----

    if Path("public").is_dir():
        app.mount("/public", StaticFiles(directory="public", html=True))
    app.mount("/", StaticFiles(directory=config.public_path, html=True))

    return app

def serve(func, config, name, port):
    import uvicorn, logging
    from dotenv import load_dotenv
    load_dotenv()
    if isinstance(config, dict):
        config = Config(**config)
    logging.getLogger("uvicorn.error").addFilter(lambda r: "0.0.0.0" not in r.getMessage())
    print(f"\n🔨 {name} => http://localhost:{port}\n")
    uvicorn.run(web(func, config), host="0.0.0.0", port=port)



# The tension for @cycls.agent is that we need both: send images/PDFs as content blocks to the model AND copy
# files to the workspace so bash/editor can operate on them. This app only needs the former.

import asyncio, base64, json, os, shutil
from dataclasses import dataclass, field
from typing import List, Optional, Union
from urllib.parse import unquote

from .app import App

COMPACT_THRESHOLD = 100_000

_MEDIA_TYPES = {
    ".jpg": "image/jpeg", ".jpeg": "image/jpeg",
    ".png": "image/png", ".gif": "image/gif", ".webp": "image/webp",
    ".pdf": "application/pdf",
}


def find_part(messages, role, ptype):
    for msg in reversed(getattr(messages, "raw", None) or messages):
        if role and msg.get("role") != role:
            continue
        for part in msg.get("parts", []) or []:
            if part.get("type") == ptype:
                return part
        if role:
            return None
    return None


def setup_workspace(context):
    user_id = context.user.id if context.user else "default"
    org_id = context.user.org_id if context.user else None
    ws = f"/workspace/{org_id}" if org_id else f"/workspace/{user_id}"
    os.makedirs(ws, exist_ok=True)
    content = context.messages.raw[-1].get("content", "")
    if not isinstance(content, list):
        return ws, content
    blocks = []
    has_media = False
    for p in content:
        if p.get("type") == "text":
            blocks.append({"type": "text", "text": p["text"]})
            continue
        if p.get("type") not in ("image", "file"):
            continue
        url = unquote(p.get("image") or p.get("file") or "")
        if not url:
            continue
        fname = os.path.basename(url)
        src = os.path.realpath(f"/workspace{url}")
        if not src.startswith("/workspace/"):
            continue
        shutil.copy(src, f"{ws}/{fname}")
        ext = os.path.splitext(fname)[1].lower()
        media_type = _MEDIA_TYPES.get(ext)
        if ext == ".pdf" and media_type:
            has_media = True
            with open(src, "rb") as f:
                data = base64.b64encode(f.read()).decode()
            blocks.append({"type": "document", "source": {"type": "base64", "media_type": media_type, "data": data}})
        elif media_type and media_type.startswith("image/"):
            has_media = True
            with open(src, "rb") as f:
                data = base64.b64encode(f.read()).decode()
            blocks.append({"type": "image", "source": {"type": "base64", "media_type": media_type, "data": data}})
        else:
            try:
                text = open(src, "r").read()
                if len(text) > 400_000:
                    text = text[:400_000] + "\n\n[... truncated ...]"
                blocks.append({"type": "text", "text": f"[File: {fname}]\n{text}\n[End of file]"})
            except (UnicodeDecodeError, ValueError):
                blocks.append({"type": "text", "text": f"[USER UPLOADED {fname}]"})
    if not has_media:
        return ws, " ".join(b["text"] for b in blocks if b.get("type") == "text")
    return ws, blocks


@dataclass
class ClaudeAgentOptions:
    workspace: str
    prompt: Union[str, list]
    model: str = "claude-sonnet-4-20250514"
    tools: List[dict] = field(default_factory=list)
    system: str = ""
    max_tokens: int = 16384
    thinking: bool = True
    session_id: Optional[str] = None


def _load_history(path):
    messages = []
    try:
        with open(path) as f:
            for i, line in enumerate(f):
                line = line.strip()
                if line:
                    messages.append(json.loads(line))
    except (FileNotFoundError, json.JSONDecodeError):
        return []
    except UnicodeDecodeError as e:
        print(f"[DEBUG] UnicodeDecodeError in {path} at line {i}: {e}")
        return messages
    for msg in messages:
        c = msg.get("content")
        if isinstance(c, list):
            for b in c:
                if isinstance(b, dict):
                    b.pop("cache_control", None)
    if messages:
        c = messages[-1].get("content")
        if isinstance(c, str):
            messages[-1]["content"] = [{"type": "text", "text": c, "cache_control": {"type": "ephemeral"}}]
        elif isinstance(c, list) and c:
            c[-1]["cache_control"] = {"type": "ephemeral"}
    return messages

def _append_history(path, messages):
    with open(path, "a") as f:
        for msg in messages:
            f.write(json.dumps(msg) + "\n")

def _rewrite_history(path, messages):
    with open(path, "w") as f:
        for msg in messages:
            f.write(json.dumps(msg) + "\n")

async def _compact(client, model, messages):
    """Summarize conversation into a compact continuation message."""
    response = await client.messages.create(
        model=model, max_tokens=8192,
        system=[{"type": "text", "text": (
            "Summarize this conversation concisely but thoroughly. Include: "
            "key decisions made, code changes with file paths, current state "
            "of work, and any pending tasks. This summary replaces the full "
            "conversation history."
        )}],
        messages=messages + [{"role": "user", "content": "Summarize our conversation so far for continuity."}],
    )
    summary = response.content[0].text
    return [
        {"role": "user", "content": (
            "This session is being continued from a previous conversation "
            "that ran out of context. The conversation is summarized below:\n\n"
            + summary
        )},
        {"role": "assistant", "content": "Understood. I have the full context from our previous conversation. How can I help?"},
    ]

def _build_tools(custom):
    tools = [
        {"type": "bash_20250124", "name": "bash"},
        {"type": "text_editor_20250728", "name": "str_replace_based_edit_tool"},
        {"type": "web_search_20250305", "name": "web_search"},
    ] + [{"type": "custom", "name": t["name"], "description": t.get("description", ""),
          "input_schema": t.get("inputSchema", t.get("input_schema", {}))} for t in custom or []]
    tools[-1]["cache_control"] = {"type": "ephemeral"}
    return tools

def _tool_result(bid, content):
    return {"type": "tool_result", "tool_use_id": bid, "content": content}

async def _exec_bash(command, cwd):
    proc = await asyncio.create_subprocess_exec(
        "bash", "-c", command,
        stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE, cwd=cwd)
    stdout, stderr = await asyncio.wait_for(proc.communicate(), timeout=300)
    out = stdout.decode(errors="replace") + (stderr.decode(errors="replace") if stderr else "")
    if len(out) > 20000:
        out = out[:10000] + "\n... (truncated) ...\n" + out[-10000:]
    return out.strip() or "(no output)"

def _exec_editor(inp, workspace):
    import pathlib
    cmd, path = inp["command"], pathlib.Path(inp["path"])
    if not path.is_absolute():
        path = pathlib.Path(workspace) / path
    path = path.resolve()
    if not str(path).startswith(os.path.realpath(workspace)):
        return f"Error: path {path} is outside workspace"
    if cmd != "create" and not path.exists():
        return f"Error: {path} does not exist"
    if cmd == "view":
        ext = path.suffix.lower()
        media_type = _MEDIA_TYPES.get(ext)
        if media_type:
            data = base64.b64encode(path.read_bytes()).decode()
            kind = "document" if not media_type.startswith("image/") else "image"
            return [{"type": kind, "source": {"type": "base64", "media_type": media_type, "data": data}}]
        try:
            lines = path.read_text().splitlines()
        except UnicodeDecodeError:
            return f"Error: {path} is a binary file"
        vr = inp.get("view_range")
        start = vr[0] if vr else 1
        if vr: lines = lines[vr[0] - 1:vr[1]]
        return "\n".join(f"{i + start:6}\t{l}" for i, l in enumerate(lines))
    if cmd == "str_replace":
        text, old = path.read_text(), inp["old_str"]
        count = text.count(old)
        if count == 0: return f"Error: old_str not found in {path}"
        if count > 1: return f"Error: old_str found {count} times, must be unique"
        path.write_text(text.replace(old, inp.get("new_str", ""), 1))
        return f"Replaced in {path}"
    if cmd == "create":
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_text(inp["file_text"])
        return f"Created {path}"
    if cmd == "insert":
        lines = path.read_text().splitlines(keepends=True)
        new = inp["new_str"].splitlines(keepends=True)
        if not new[-1:] or not new[-1].endswith("\n"): new.append("\n")
        pos = inp["insert_line"]
        lines[pos:pos] = new
        path.write_text("".join(lines))
        return f"Inserted at line {pos} in {path}"
    return f"Error: unknown command {cmd}"


async def ClaudeAgent(*, options):
    """Run one Claude turn. Async generator yielding streaming components."""
    import anthropic, uuid

    ws = options.workspace
    client = anthropic.AsyncAnthropic()

    sid = options.session_id or str(uuid.uuid4())
    if not options.session_id:
        yield {"type": "session_id", "session_id": sid}

    history_path = os.path.join(ws, ".cycls", f"{sid}.jsonl")
    os.makedirs(os.path.dirname(history_path), exist_ok=True)
    messages = _load_history(history_path)
    loaded_count = len(messages)
    messages.append({"role": "user", "content": options.prompt})

    kwargs = {
        "model": options.model, "max_tokens": options.max_tokens,
        "tools": _build_tools(options.tools), "messages": messages,
    }
    if options.system:
        kwargs["system"] = [{"type": "text", "text": options.system, "cache_control": {"type": "ephemeral"}}]
    if options.thinking:
        kwargs["thinking"] = {"type": "adaptive"}

    total_in = total_out = cache_read = cache_create = 0

    try:
        while True:
            thinking_text = ""
            search_idx, search_query = None, ""

            async with client.messages.stream(**kwargs) as stream:
                async for ev in stream:
                    if ev.type == "content_block_start":
                        b = ev.content_block
                        if b.type == "server_tool_use" and b.name == "web_search":
                            search_idx, search_query = ev.index, ""
                    elif ev.type == "content_block_delta":
                        if ev.delta.type == "thinking_delta":
                            thinking_text += ev.delta.thinking
                        elif ev.delta.type == "text_delta":
                            yield ev.delta.text
                        elif ev.delta.type == "input_json_delta" and ev.index == search_idx:
                            search_query += ev.delta.partial_json
                    elif ev.type == "content_block_stop":
                        if thinking_text:
                            yield {"type": "thinking", "thinking": thinking_text}
                            thinking_text = ""
                        if ev.index == search_idx:
                            try: q = json.loads(search_query).get("query", "")
                            except Exception: q = ""
                            yield {"type": "step", "step": f'Web Search("{q}")' if q else "Web Search"}
                            search_idx = None
                response = await stream.get_final_message()

            total_in += response.usage.input_tokens
            total_out += response.usage.output_tokens
            cache_read += response.usage.cache_read_input_tokens or 0
            cache_create += response.usage.cache_creation_input_tokens or 0

            content = [b.model_dump(exclude_none=True) for b in response.content]
            messages.append({"role": "assistant", "content": content})
            tool_blocks = [b for b in response.content if b.type == "tool_use"]

            if response.stop_reason != "tool_use":
                break

            # Yield step indicators, then execute tools in parallel
            for block in tool_blocks:
                name, inp = block.name, block.input
                if name == "bash":
                    yield {"type": "step", "step": f"Bash({inp.get('command', '')[:60]})"}
                elif name == "str_replace_based_edit_tool":
                    verb = "Editing" if inp.get("command") in ("str_replace", "create", "insert") else "Viewing"
                    yield {"type": "step", "step": f"{verb} {inp.get('path', 'file')}"}
                else:
                    yield {"type": "tool_call", "tool": name, "args": inp}

            async def _run_tool(block):
                name, inp = block.name, block.input
                if name == "bash":
                    return await _exec_bash(inp.get("command", ""), ws)
                elif name == "str_replace_based_edit_tool":
                    return _exec_editor(inp, ws)
                return f"{name} rendered successfully"

            if len(tool_blocks) > 1:
                print(f"[PARALLEL] Running {len(tool_blocks)} tools concurrently")
            outputs = await asyncio.gather(*[_run_tool(b) for b in tool_blocks])
            results = []
            for block, out in zip(tool_blocks, outputs):
                if block.name == "bash":
                    yield {"type": "step_data", "data": out}
                results.append(_tool_result(block.id, out))
            messages.append({"role": "user", "content": results})

    except Exception as e:
        import traceback
        print(f"[DEBUG] ClaudeAgent error: {e}\n{traceback.format_exc()}")
        yield {"type": "callout", "callout": str(e), "style": "error"}

    new_messages = messages[loaded_count:]
    if total_in > COMPACT_THRESHOLD and messages:
        try:
            yield {"type": "step", "step": "Compacting context..."}
            messages = await _compact(client, options.model, messages)
            _rewrite_history(history_path, messages)
        except Exception:
            if new_messages and new_messages[-1].get("role") == "assistant":
                _append_history(history_path, new_messages)
    elif new_messages and new_messages[-1].get("role") == "assistant":
        _append_history(history_path, new_messages)
    if total_in:
        yield {"type": "usage", "usage": {"tokenUsage": {"total": {
            "inputTokens": total_in, "outputTokens": total_out,
            "cachedInputTokens": cache_read, "cacheCreationTokens": cache_create,
        }}}}


class Agent(App):
    def __init__(self, func, name, pip=None, **kwargs):
        super().__init__(func=func, name=name, pip=["anthropic", *(pip or [])], **kwargs)

def agent(name=None, **kwargs):
    def decorator(func):
        return Agent(func=func, name=name or func.__name__, **kwargs)
    return decorator



import contextlib
import docker
import cloudpickle
import tempfile
import hashlib
import json
import os
import sys
import shutil
from pathlib import Path
import tarfile
from dotenv import load_dotenv

load_dotenv()

os.environ["DOCKER_BUILDKIT"] = "1"

ENTRYPOINT_PY = '''import sys
sys.path.insert(0, '/app')
import cloudpickle
with open("/app/function.pkl", "rb") as f:
    func, args, kwargs = cloudpickle.load(f)
func(*args, **kwargs)
'''

RUNNER_PY = '''import sys
sys.path.insert(0, '/app')
import cloudpickle
import traceback
from pathlib import Path

io_dir = Path(sys.argv[1])
payload_path = io_dir / "payload.pkl"
result_path = io_dir / "result.pkl"

try:
    with open(payload_path, "rb") as f:
        func, args, kwargs = cloudpickle.load(f)
    result = func(*args, **kwargs)
    with open(result_path, "wb") as f:
        cloudpickle.dump(result, f)
except Exception:
    traceback.print_exc()
    sys.exit(1)
'''

# Module-level configuration
api_key = None
base_url = None

def _get_api_key():
    import sys
    cycls_pkg = sys.modules.get('cycls')
    return api_key or (cycls_pkg and cycls_pkg.__dict__.get('api_key')) or os.getenv("CYCLS_API_KEY")

def _get_base_url():
    import sys
    cycls_pkg = sys.modules.get('cycls')
    return base_url or (cycls_pkg and cycls_pkg.__dict__.get('base_url')) or os.getenv("CYCLS_BASE_URL")

def _hash_path(path_str: str) -> str:
    h = hashlib.sha256()
    p = Path(path_str)
    if p.is_file():
        with p.open('rb') as f:
            while chunk := f.read(65536):
                h.update(chunk)
    elif p.is_dir():
        for root, dirs, files in os.walk(p, topdown=True):
            dirs.sort()
            files.sort()
            for name in files:
                filepath = Path(root) / name
                h.update(str(filepath.relative_to(p)).encode())
                with filepath.open('rb') as f:
                    while chunk := f.read(65536):
                        h.update(chunk)
    return h.hexdigest()


def _copy_path(src_path: Path, dest_path: Path):
    if src_path.is_dir():
        shutil.copytree(src_path, dest_path, dirs_exist_ok=True)
    else:
        dest_path.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy(src_path, dest_path)


class Function:
    """Executes functions in Docker containers."""

    def __init__(self, func, name, python_version=None, pip=None, apt=None,
                 run_commands=None, copy=None, base_url=None, api_key=None, force_rebuild=False):
        self.func = func
        self.name = name.replace('_', '-')
        self.python_version = python_version or f"{sys.version_info.major}.{sys.version_info.minor}"
        self.base_image = f"python:{self.python_version}-slim"
        self.apt = sorted(apt or [])
        self.run_commands = list(run_commands or [])
        self.copy = {f: f for f in copy} if isinstance(copy, list) else (copy or {})
        self._base_url = base_url
        self._api_key = api_key
        self.pip = sorted(set(pip or []) | {"cloudpickle"})
        self.force_rebuild = force_rebuild

        self.image_prefix = f"cycls/{self.name}"
        self.managed_label = "cycls.function"
        self._docker_client = None
        self._container = None

    @property
    def api_key(self):
        return self._api_key or _get_api_key()

    @property
    def base_url(self):
        return self._base_url or _get_base_url() or "https://api.cycls.ai"

    @property
    def docker_client(self):
        if self._docker_client is None:
            try:
                print("Initializing Docker client...")
                client = docker.from_env()
                client.ping()
                self._docker_client = client
            except docker.errors.DockerException:
                print("\nError: Docker is not running or is not installed.")
                print("Please start the Docker daemon and try again.")
                sys.exit(1)
        return self._docker_client

    def _perform_auto_cleanup(self, keep_tag=None):
        try:
            current_id = self._container.id if self._container else None
            for container in self.docker_client.containers.list(all=True, filters={"label": self.managed_label}):
                if container.id != current_id:
                    container.remove(force=True)

            cleaned = 0
            for image in self.docker_client.images.list(filters={"label": self.managed_label}):
                is_deploy = any(":deploy-" in t for t in image.tags)
                is_current = keep_tag and keep_tag in image.tags
                if not is_deploy and not is_current:
                    self.docker_client.images.remove(image.id, force=True)
                    cleaned += 1
            if cleaned:
                print(f"Cleaned up {cleaned} old dev image(s).")
        except Exception as e:
            print(f"Warning: cleanup error: {e}")

    def _image_tag(self, extra_parts=None) -> str:
        parts = [self.base_image, self.python_version, "".join(self.pip),
                 "".join(self.apt), "".join(self.run_commands)]
        for src, dst in sorted(self.copy.items()):
            if not Path(src).exists():
                raise FileNotFoundError(f"Path in 'copy' not found: {src}")
            parts.append(f"{src}>{dst}:{_hash_path(src)}")
        if extra_parts:
            parts.extend(extra_parts)
        return f"{self.image_prefix}:{hashlib.sha256(''.join(parts).encode()).hexdigest()[:16]}"

    def _dockerfile_preamble(self) -> str:
        lines = [
            f"FROM {self.base_image}",
            "ENV PIP_ROOT_USER_ACTION=ignore PYTHONUNBUFFERED=1",
            "WORKDIR /app",
            "RUN pip install uv",
        ]

        if self.apt:
            lines.append(f"RUN apt-get update && apt-get install -y --no-install-recommends {' '.join(self.apt)}")

        if self.pip:
            lines.append(f"RUN uv pip install --system --no-cache {' '.join(self.pip)}")

        for cmd in self.run_commands:
            lines.append(f"RUN {cmd}")

        for dst in self.copy.values():
            lines.append(f"COPY context_files/{dst} /app/{dst}")

        return "\n".join(lines)

    def _dockerfile_local(self) -> str:
        return f"""{self._dockerfile_preamble()}
COPY runner.py /runner.py
ENTRYPOINT ["python", "/runner.py", "/io"]
"""

    def _dockerfile_deploy(self, port: int) -> str:
        return f"""{self._dockerfile_preamble()}
COPY function.pkl /app/function.pkl
COPY entrypoint.py /app/entrypoint.py
EXPOSE {port}
CMD ["python", "entrypoint.py"]
"""

    def _copy_user_files(self, workdir: Path):
        context_files_dir = workdir / "context_files"
        context_files_dir.mkdir()
        for src, dst in self.copy.items():
            _copy_path(Path(src).resolve(), context_files_dir / dst)

    def _build_image(self, tag: str, workdir: Path) -> str:
        print("--- Docker Build Logs ---")
        force = os.environ.get('_CYCLS_FORCE_REBUILD') == '1' if os.environ.get('_CYCLS_WATCH') else self.force_rebuild
        try:
            for chunk in self.docker_client.api.build(
                path=str(workdir), tag=tag, forcerm=True, decode=True,
                nocache=force, labels={self.managed_label: "true"}
            ):
                if 'stream' in chunk:
                    print(chunk['stream'].strip())
            print("-------------------------")
            print(f"Image built: {tag}")
            return tag
        except docker.errors.BuildError as e:
            print(f"\nDocker build failed: {e}")
            raise

    def _ensure_local_image(self) -> str:
        tag = self._image_tag(extra_parts=["local-v1"])
        force = os.environ.get('_CYCLS_FORCE_REBUILD') == '1' if os.environ.get('_CYCLS_WATCH') else self.force_rebuild
        if not force:
            try:
                self.docker_client.images.get(tag)
                print(f"Found cached image: {tag}")
                return tag
            except docker.errors.ImageNotFound:
                pass
        print(f"Building new image: {tag}")

        with tempfile.TemporaryDirectory() as tmpdir:
            workdir = Path(tmpdir)
            self._copy_user_files(workdir)
            (workdir / "Dockerfile").write_text(self._dockerfile_local())
            (workdir / "runner.py").write_text(RUNNER_PY)
            return self._build_image(tag, workdir)

    def _cleanup_container(self):
        if getattr(self, '_container', None):
            try:
                self._container.stop(timeout=3)
                self._container.remove()
            except docker.errors.NotFound:
                pass
            except docker.errors.APIError:
                pass
            self._container = None

    @contextlib.contextmanager
    def runner(self, *args, **kwargs):
        service_port = kwargs.get('port')
        tag = self._ensure_local_image()
        self._perform_auto_cleanup(keep_tag=tag)

        ports = {f'{service_port}/tcp': service_port} if service_port else None

        with tempfile.TemporaryDirectory() as io_dir:
            io_path = Path(io_dir)
            payload_path = io_path / "payload.pkl"
            result_path = io_path / "result.pkl"

            with open(payload_path, 'wb') as f:
                cloudpickle.dump((self.func, args, kwargs), f)

            try:
                self._container = self.docker_client.containers.create(
                    image=tag,
                    volumes={str(io_path): {'bind': '/io', 'mode': 'rw'}},
                    ports=ports,
                    labels={self.managed_label: "true"}
                )
                self._container.start()
                yield self._container, result_path
            finally:
                self._cleanup_container()

    def run(self, *args, **kwargs):
        service_port = kwargs.get('port')
        print(f"Running '{self.name}'...")

        try:
            with self.runner(*args, **kwargs) as (container, result_path):
                print("--- Container Logs ---")
                for chunk in container.logs(stream=True, follow=True):
                    print(chunk.decode(), end='')
                print("----------------------")

                status = container.wait()
                if status['StatusCode'] != 0:
                    print(f"Error: Container exited with code {status['StatusCode']}")
                    return None

                if service_port:
                    return None

                if result_path.exists():
                    with open(result_path, 'rb') as f:
                        return cloudpickle.load(f)
                else:
                    print("Error: Result file not found")
                    return None

        except KeyboardInterrupt:
            print("\n----------------------")
            print("Stopping...")
            return None
        except Exception as e:
            print(f"Error: {e}")
            return None

    def watch(self, *args, **kwargs):
        if os.environ.get('_CYCLS_WATCH'):
            return self.run(*args, **kwargs)

        try:
            from watchfiles import watch as watchfiles_watch
        except ImportError:
            print("watchfiles not installed. pip install watchfiles")
            return self.run(*args, **kwargs)

        import subprocess

        script = Path(sys.argv[0]).resolve()
        watch_paths = [script] + [Path(p).resolve() for p in self.copy if Path(p).exists()]

        print(f"Watching: {[p.name for p in watch_paths]}\n")

        first_run = True
        while True:
            env = {**os.environ, '_CYCLS_WATCH': '1'}
            # Force rebuild only on first run, then use cache for subsequent reloads
            env['_CYCLS_FORCE_REBUILD'] = '1' if (self.force_rebuild and first_run) else '0'

            proc = subprocess.Popen([sys.executable, str(script)], env=env)
            first_run = False
            try:
                for changes in watchfiles_watch(*watch_paths):
                    print(f"\nChanged: {[Path(c[1]).name for c in changes]}")
                    break
                proc.terminate()
                proc.wait(timeout=3)
            except subprocess.TimeoutExpired:
                proc.kill()
            except KeyboardInterrupt:
                proc.terminate()
                proc.wait(timeout=3)
                return

    def _prepare_deploy_context(self, workdir: Path, port: int, args=(), kwargs=None):
        kwargs = kwargs or {}
        kwargs['port'] = port
        self._copy_user_files(workdir)
        (workdir / "Dockerfile").write_text(self._dockerfile_deploy(port))
        (workdir / "entrypoint.py").write_text(ENTRYPOINT_PY)
        with open(workdir / "function.pkl", "wb") as f:
            cloudpickle.dump((self.func, args, kwargs), f)

    def build(self, *args, **kwargs):
        port = kwargs.pop('port', 8080)
        payload = cloudpickle.dumps((self.func, args, {**kwargs, 'port': port}))
        tag = f"{self.image_prefix}:deploy-{hashlib.sha256(payload).hexdigest()[:16]}"

        try:
            self.docker_client.images.get(tag)
            print(f"Found cached image: {tag}")
            return tag
        except docker.errors.ImageNotFound:
            print(f"Building: {tag}")

        with tempfile.TemporaryDirectory() as tmpdir:
            workdir = Path(tmpdir)
            self._prepare_deploy_context(workdir, port, args, kwargs)
            self._build_image(tag, workdir)
            print(f"Run: docker run --rm -p {port}:{port} {tag}")
            return tag

    def deploy(self, *args, **kwargs):
        import requests

        base_url = self.base_url
        port = kwargs.pop('port', 8080)
        memory = kwargs.pop('memory', '1Gi')

        # Check name availability before uploading
        print(f"Checking '{self.name}'...")
        try:
            check_resp = requests.get(
                f"{base_url}/v1/deployment/check-name",
                params={"name": self.name},
                headers={"X-API-Key": self.api_key},
                timeout=30,
            )
            if check_resp.status_code == 401:
                print("Error: Invalid API key")
                return None
            check_resp.raise_for_status()
            check_data = check_resp.json()
            if not check_data.get("available"):
                print(f"Error: {check_data.get('reason', 'Name unavailable')}")
                return None
        except requests.exceptions.RequestException as e:
            print(f"Error checking name: {e}")
            return None

        print(f"Deploying '{self.name}'...")

        payload = cloudpickle.dumps((self.func, args, {**kwargs, 'port': port}))
        archive_name = f"{self.name}-{hashlib.sha256(payload).hexdigest()[:16]}.tar.gz"

        with tempfile.TemporaryDirectory() as tmpdir:
            workdir = Path(tmpdir)
            self._prepare_deploy_context(workdir, port, args, kwargs)

            archive_path = workdir / archive_name
            with tarfile.open(archive_path, "w:gz") as tar:
                for f in workdir.glob("**/*"):
                    if f.is_file() and f != archive_path:
                        tar.add(f, arcname=f.relative_to(workdir))

            print("Uploading...")
            with open(archive_path, 'rb') as f:
                response = requests.post(
                    f"{base_url}/v1/deploy",
                    data={"function_name": self.name, "port": port, "memory": memory},
                    files={'source_archive': (archive_name, f, 'application/gzip')},
                    headers={"X-API-Key": self.api_key},
                    timeout=9000,
                    stream=True,
                )

            if not response.ok:
                print(f"Deploy failed: {response.status_code}")
                try:
                    print(f"  {response.json()['detail']}")
                except (json.JSONDecodeError, KeyError):
                    print(f"  {response.text}")
                return None

            # Parse NDJSON stream
            url = None
            for line in response.iter_lines(decode_unicode=True):
                if line:
                    event = json.loads(line)
                    status = event.get("status", "")
                    msg = event.get("message", "")
                    print(f"  [{status}] {msg}")
                    if status == "DONE":
                        url = event.get("url")
                        print(f"Deployed: {url}")
                    elif status == "ERROR":
                        return None
            return url

    def __del__(self):
        self._cleanup_container()


def function(name=None, **kwargs):
    """Decorator that transforms a Python function into a containerized Function."""
    def decorator(func):
        return Function(func, name or func.__name__, **kwargs, base_url=_get_base_url(), api_key=_get_api_key())
    return decorator
